(*
 * LANGUAGE    : ANS Forth with DFW extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : 32-bit absolute relocater
 * CATEGORY    : Tool
 * AUTHOR      : Marcel Hendrix 
 * LAST CHANGE : September 26, 1995, Marcel Hendrix 
 *)



        NEEDS -miscutil
	NEEDS -streams

        REVISION -relocate " Relocate TASM files Version 1.00 "

	PRIVATES


DOC
(*
 Purpose:
 --------
 Relocate to an absolute address (typically $40005000) .exe files assembled 
 by TASM and linked by TLINK. TLINK32 can/will not do this (it tries to 
 write an exe file with $40005000 zero's before the first code byte).

 The resulting .exe file can be used together with the MMURTL OS (strip off
 the symbol table and change everything to DB statements).

 Algorithm
 ---------
 Open two (exe)files f1 and f2 that are assembled for addresses org1 and org2.
 Open a new file f3 for output.

 Step through f1 and f2 a byte at a time and read the 32-bit numbers found at 
 that offset (offs). 
 Check the difference between the numbers. Is it org2-org1?
   Yes --> the value is discarded, neworg+delta is written to the file and the
           pipeline is refilled with four new bytes
   No  --> Shift the pipeline. Write the byte shifted out to f3 and read one 
           new byte to refill.

 Remark: We assume that there are no 16-bit or 8-bit addresses in the file.
 Remark: When EOF is detected there can be up to 3 unwritten bytes in the 
	 pipeline.
*)
ENDDOC

STREAM <in1>	PRIVATE 
STREAM <in2>	PRIVATE
STREAM <out>	PRIVATE

CREATE ilong1 PRIVATE	5 CHARS ALLOT	ilong1 char+     =: ilong1+1 PRIVATE
					ilong1 4 chars + =: ilong1+4 PRIVATE
CREATE ilong2 PRIVATE	5 CHARS ALLOT	ilong2 char+     =: ilong2+1 PRIVATE
					ilong2 4 chars + =: ilong2+4 PRIVATE
	
	0 VALUE cnt 		PRIVATE
    $1000 VALUE org2-org1 	PRIVATE
    $1000 VALUE org1		PRIVATE
$40005000 VALUE org		PRIVATE

: ONE@	( -- bool )
	ilong1+1 @  ilong1 !
	SGETCH <in1> DUP ilong1+4 C!
	ilong2+1 @  ilong2 !
	SGETCH <in2> DUP ilong2+4 C!
	EOF = SWAP 
	EOF = OR ;p

: FOUR@	 ( -- )   	4 0 DO one@ drop LOOP CLEAR cnt ;p
: ONE!	 ( -- )		ilong1 C@ SPUTCH <out> 1 +TO cnt ;p
: BYTES! ( n u -- )	0 ?DO DUP SPUTCH <out> 8 RSHIFT LOOP DROP ;p
: ORG+!	 ( -- )		ilong1+1 @ org1 -  org + 4 bytes! ;p

: OLD!	 ( -- )   	ilong2+1 @ ilong1+1 @ -  org2-org1 
			 = IF org+! 
			 ELSE ilong1+1 @ 3 bytes! 
			ENDIF ;p

: .PARAMETERS ( -- )
	CR ." ORG of file1 is " org1 H. 
	CR ." ORG of file2 is " org1 org2-org1 + H.
	CR ." (absolute) address of file3 will be set to " org H. ;

: SET-PARAMETERS ( org1 org2 neworg -- )
	TO org OVER - TO org2-org1  TO org1 ;

: RELOCATE ( c-addr1 u1 c-addr2 u2 c-addr2 u3 -- )
	$OPEN-OUTPUT <out>
	$OPEN-INPUT  <in2>
	$OPEN-INPUT  <in1>
	four@
	BEGIN
	  ilong2+1 @ ilong1+1 @ - 
	  org2-org1 = IF org+! four@ FALSE
	            ELSE one@ ( bool) one!  
	           ENDIF 
	UNTIL
	old!
	SCLOSE-INPUT  <in1> 
	SCLOSE-INPUT  <in2> 
	SCLOSE-OUTPUT <out> ;

: SPECIAL ( -- ) \ MMURTL
	$1000 $2000 $40005000 SET-PARAMETERS
	S" eforth1.exe" S" eforth2.exe" S" 3.exe" RELOCATE 
	EVAL" include exe2out.frt special bye" ;

: SPECIAL2 ( -- ) \ Linux
	$1000 $2000 $804B000 SET-PARAMETERS
	CR ." Relocating for $804B000..."
	S" eforth1.exe" S" eforth2.exe" S" 3.exe" RELOCATE 
	EVAL" include exe2out.frt special2 bye" ;

:ABOUT	CR ." Try: S" &" EMIT ."  name1" &" EMIT 
		   ."  S" &" EMIT ."  name2"   &" EMIT 
		   ."  S" &" EMIT ."  outname" &" EMIT 
		   ."  RELOCATE" 
	CR ."      <org1> <org2> <neworg> SET-PARAMETERS"
	CR ."      .PARAMETERS " 
	CR
	CR ." Current Parameters: " .PARAMETERS ;

		DEPRIVE
                .ABOUT -relocate CR

                              (* End of Source *)